from retejo._internal.interfaces.client import AsyncClient, SyncClient
from retejo._internal.interfaces.factory import Factory
from retejo._internal.interfaces.request_context_builder import (
    RequestContext,
    RequestContextBuilder,
)
from retejo._internal.interfaces.sendable_method import AsyncSendableMethod, SyncSendableMethod
from retejo._internal.interfaces.sendable_request import (
    AsyncSendableRequest,
    Request,
    Response,
    SyncSendableRequest,
)

__all__ = [
    "AsyncClient",
    "AsyncSendableMethod",
    "AsyncSendableRequest",
    "Factory",
    "Request",
    "RequestContext",
    "RequestContextBuilder",
    "Response",
    "SyncClient",
    "SyncSendableMethod",
    "SyncSendableRequest",
]



from dataclasses import dataclass
from typing import override


@dataclass
class RetejoError(Exception):
    pass


@dataclass
class MalformedResponseError(RetejoError):
    pass


@dataclass
class ClientLibraryError(RetejoError):
    pass


@dataclass
class ClientError(RetejoError, RuntimeError):
    status_code: int

    @override
    def __str__(self) -> str:
        return f"Client error with {self.status_code!r} error code"


@dataclass
class ServerError(RetejoError, RuntimeError):
    status_code: int

    @override
    def __str__(self) -> str:
        return f"Server error with {self.status_code!r} error code"



from collections import defaultdict
from collections.abc import Mapping, MutableMapping, Sequence
from dataclasses import Field, dataclass, fields as get_fields
from typing import TYPE_CHECKING, Any, NotRequired, TypedDict

from retejo._internal.markers.base import BaseMarker, get_marker_type
from retejo._internal.markers.omitted import is_omittable

if TYPE_CHECKING:
    from retejo._internal.method.method import Method


@dataclass(slots=True, frozen=True)
class MethodContext:
    fields: Mapping[type[BaseMarker], Sequence[Field[Any]]]
    types: Mapping[type[BaseMarker], type[Any] | None]


def create_method_context(method_tp: "type[Method[Any]]") -> MethodContext:
    fields = get_marker_fields(get_fields(method_tp))

    types = {}
    for marker_tp, marker_fields in fields.items():
        types[marker_tp] = make_typed_dict_for_marker(
            method_tp=method_tp,
            marker_tp=marker_tp,
            fields=marker_fields,
        )

    return MethodContext(
        types=types,
        fields=fields,
    )


def get_marker_fields(
    fields: Sequence[Field[Any]],
) -> Mapping[type[BaseMarker], Sequence[Field[Any]]]:
    result = defaultdict(list)

    for field in fields:
        marker_tp = get_marker_type(field.type)
        if marker_tp is not None:
            result[marker_tp].append(field)

    return result


def make_typed_dict_for_marker(
    method_tp: "type[Method[Any]]",
    marker_tp: type[BaseMarker],
    fields: Sequence[Field[Any]],
) -> Any | None:
    if not fields:
        return None

    name = f"{method_tp.__name__}{marker_tp.__name__}Type"

    fields_tp: MutableMapping[str, Any] = {}
    for field in fields:
        if is_omittable(field.type):
            fields_tp[field.name] = NotRequired[field.type]
        else:
            fields_tp[field.name] = field.type

    return TypedDict(name, fields_tp)  # type: ignore[operator]



from abc import abstractmethod
from collections.abc import Mapping, MutableMapping
from types import NoneType
from typing import Any, Protocol, override

from retejo.errors import ClientError, ServerError
from retejo.interfaces import (
    AsyncClient,
    Factory,
    Request,
    RequestContextBuilder,
    Response,
    SyncClient,
)
from retejo.markers import (
    BaseMarker,
    BodyMarker,
    HeaderMarker,
    QueryParamMarker,
    UrlVarMarker,
)
from retejo.method import Method
from retejo.request_context_builder import SimpleRequestContextBuilder

type MarkersFactorties = MutableMapping[type[BaseMarker], Factory]


class BaseClient(Protocol):
    _response_factory: Factory
    _request_context_builder: RequestContextBuilder
    _markers_factories: MarkersFactorties

    def __init__(self) -> None:
        self._markers_factories = self.init_markers_factories()
        self._response_factory = self.init_response_factory()
        self._request_context_builder = self.init_request_context_builder()

    @abstractmethod
    def init_markers_factories(self) -> MarkersFactorties:
        raise NotImplementedError

    @abstractmethod
    def init_response_factory(self) -> Factory:
        raise NotImplementedError

    def init_request_context_builder(self) -> RequestContextBuilder:
        return SimpleRequestContextBuilder(self._markers_factories)

    def _method_to_request(self, method: Method[Any]) -> Request:
        request_context = self._request_context_builder.build(method)

        url_vars = request_context.get(UrlVarMarker)

        if url_vars is None:  # noqa: SIM108
            url = method.__url__
        else:
            url = method.__url__.format_map(url_vars)

        return Request(
            url=url,
            http_method=method.__method__,
            body=request_context.get(BodyMarker),
            headers=request_context.get(HeaderMarker),
            query_params=request_context.get(QueryParamMarker),
            context=request_context,
        )

    def _load_method_returning[T](
        self,
        response: Mapping[str, Any],
        method_returning_tp: type[T],
    ) -> T:
        if method_returning_tp is NoneType:
            return None  # type: ignore[return-value]

        return self._response_factory.load(response, method_returning_tp)


class SyncBaseClient(BaseClient, SyncClient):
    @override
    def _handle_response(self, response: Response) -> None:
        if response.status_code >= 400:
            self._handle_error_response(response)

    @override
    def _handle_error_response(self, response: Response) -> None:
        if 400 <= response.status_code < 500:
            raise ClientError(response.status_code)
        else:
            raise ServerError(response.status_code)

    @override
    def send_method[T](
        self,
        method: Method[T],
    ) -> T:
        request = self._method_to_request(method)
        response = self.send_request(request)

        self._handle_response(response)

        return self._load_method_returning(
            response=response.data,
            method_returning_tp=method.__returning__,
        )


class AsyncBaseClient(BaseClient, AsyncClient):
    @override
    async def _handle_response(self, response: Response) -> None:
        if response.status_code >= 400:
            await self._handle_error_response(response)

    @override
    async def _handle_error_response(self, response: Response) -> None:
        if 400 <= response.status_code < 500:
            raise ClientError(response.status_code)
        else:
            raise ServerError(response.status_code)

    @override
    async def send_method[T](
        self,
        method: Method[T],
    ) -> T:
        request = self._method_to_request(method)
        response = await self.send_request(request)

        await self._handle_response(response)

        return self._load_method_returning(
            response=response.data,
            method_returning_tp=method.__returning__,
        )



from .aiohttp import AiohttpBaseClient
from .base import AsyncBaseClient, BaseClient, SyncBaseClient
from .requests import RequestsBaseClient

__all__ = [
    "AiohttpBaseClient",
    "AsyncBaseClient",
    "BaseClient",
    "RequestsBaseClient",
    "SyncBaseClient",
]



from collections.abc import Mapping
from typing import Any, override

from retejo.interfaces import (
    Factory,
    RequestContext,
    RequestContextBuilder,
)
from retejo.markers import BaseMarker, is_omitted
from retejo.method import Method


class SimpleRequestContextBuilder(RequestContextBuilder):
    _markers_factories: Mapping[type[BaseMarker], Factory]

    def __init__(self, markers_factories: Mapping[type[BaseMarker], Factory]) -> None:
        self._markers_factories = markers_factories

    @override
    def build(self, method: Method[Any]) -> RequestContext:
        context = method.__context__

        result: dict[type[BaseMarker], Any] = {}

        for marker_tp, marker_fields in context.fields.items():
            data = {}

            for marker_filed in marker_fields:
                method_attr_value = getattr(method, marker_filed.name)
                if not is_omitted(method_attr_value):
                    data[marker_filed.name] = method_attr_value

            factory = self._markers_factories.get(marker_tp)
            if factory is not None:
                tp = context.types[marker_tp]
                if tp is not None:
                    data = factory.dump(data, tp)

            result[marker_tp] = data

        return result



import urllib.parse
from collections.abc import Mapping
from json import JSONDecodeError
from typing import Any, override

from requests import RequestException, Session

from retejo.errors import ClientLibraryError, MalformedResponseError
from retejo.integrations.common.base import SyncBaseClient
from retejo.interfaces import Request, Response


class RequestsBaseClient(SyncBaseClient):
    _base_url: str
    _session: Session

    def __init__(
        self,
        base_url: str,
        session: Session | None = None,
        cookies: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        super().__init__()

        self._base_url = base_url

        if session is None:
            self._session = Session()
        else:
            self._session = session

        if headers is not None:
            self._session.headers.update(headers)
        if cookies is not None:
            self._session.cookies.update(cookies)

    @override
    def send_request(
        self,
        request: Request,
    ) -> Response:
        response = self._session.request(
            method=request.http_method,
            url=urllib.parse.urljoin(self._base_url, request.url),
            params=request.query_params,
            json=request.body,
            headers=request.headers,
        )

        try:
            data = response.json()
        except RequestException as e:
            raise ClientLibraryError from e
        except JSONDecodeError as e:
            raise MalformedResponseError from e

        return Response(
            data=data,
            status_code=response.status_code,
        )

    @override
    def close(self) -> None:
        self._session.close()



from .base import is_marker
from .body import Body, is_body
from .file import File, is_file
from .header import Header, is_header
from .omitted import Omittable, Omitted, is_omittable, is_omitted
from .query_param import QueryParam, is_query_param
from .url_var import UrlVar, is_url_var

__all__ = (
    "Body",
    "File",
    "Header",
    "Omittable",
    "Omitted",
    "QueryParam",
    "UrlVar",
    "is_body",
    "is_file",
    "is_header",
    "is_marker",
    "is_omittable",
    "is_omitted",
    "is_query_param",
    "is_url_var",
)



from collections.abc import MutableMapping
from typing import Any, ClassVar, override


class SingletonMeta(type):
    _instances: ClassVar[MutableMapping[Any, Any]] = {}

    @override
    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)

        return cls._instances[cls]



from typing import TYPE_CHECKING, Annotated, Any, TypeVar

from retejo._internal.markers.base import BaseMarker, is_marker_factory


class BodyMarker(BaseMarker):
    pass


T = TypeVar("T")

if TYPE_CHECKING:
    type Body[T] = T
else:

    class Body:
        def __class_getitem__(cls, item: Any) -> Any:
            return Annotated[item, BodyMarker()]


is_body = is_marker_factory(BodyMarker)



from typing import Any, Protocol, overload, runtime_checkable


@runtime_checkable
class Factory(Protocol):
    @overload
    def load[T](self, data: Any, tp: type[T], /) -> T: ...

    @overload
    def load(self, data: Any, tp: Any, /) -> Any: ...

    def load(self, data: Any, tp: Any, /) -> Any: ...

    @overload
    def dump[T](self, data: T, tp: type[T], /) -> Any: ...

    @overload
    def dump(self, data: Any, tp: Any | None = None, /) -> Any: ...

    def dump(self, data: Any, tp: Any | None = None, /) -> Any: ...



from typing import TYPE_CHECKING, Annotated, Any

from retejo._internal.markers.base import BaseMarker, is_marker_factory


class UrlVarMarker(BaseMarker):
    pass


if TYPE_CHECKING:
    type UrlVar[T] = T
else:

    class UrlVar:
        def __class_getitem__(cls, item: Any) -> Any:
            return Annotated[item, UrlVarMarker()]


is_url_var = is_marker_factory(UrlVarMarker)



from abc import abstractmethod
from collections.abc import Mapping
from typing import Any, Protocol, runtime_checkable

from retejo._internal.markers.base import BaseMarker
from retejo.method import Method

type RequestContext = Mapping[type[BaseMarker], Mapping[str, Any]]


@runtime_checkable
class RequestContextBuilder(Protocol):
    @abstractmethod
    def build(self, method: Method[Any]) -> RequestContext:
        raise NotImplementedError



from collections.abc import Callable
from typing import Annotated, Any, TypeGuard, get_origin

from retejo._internal.singlton import SingletonMeta


class BaseMarker(metaclass=SingletonMeta):
    pass


def is_marker_factory[T: BaseMarker](marker: type[T]) -> Callable[[Any], TypeGuard[T]]:
    def wrapper(obj: Any) -> TypeGuard[T]:
        if get_origin(obj) is Annotated:
            return isinstance(obj.__metadata__[0], marker)
        return False

    return wrapper


def get_marker_type(obj: Any) -> type[BaseMarker] | None:
    if get_origin(obj) is Annotated:
        args = obj.__metadata__
        return type(next(arg for arg in args if isinstance(arg, BaseMarker)))

    return None


is_marker = is_marker_factory(BaseMarker)



from collections.abc import Mapping
from json import JSONDecodeError
from typing import Any, override

from aiohttp import ClientError, ClientSession

from retejo.errors import ClientLibraryError, MalformedResponseError
from retejo.integrations.common.base import AsyncBaseClient
from retejo.interfaces import Request, Response


class AiohttpBaseClient(AsyncBaseClient):
    _session: ClientSession

    def __init__(
        self,
        base_url: str,
        session: ClientSession | None = None,
        cookies: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        super().__init__()

        if session is None:
            self._session = ClientSession(base_url)
        else:
            self._session = session

        if headers is not None:
            self._session.headers.update(headers)
        if cookies is not None:
            self._session.cookie_jar.update_cookies(cookies)

    @override
    async def send_request(
        self,
        request: Request,
    ) -> Response:
        async with self._session.request(
            method=request.http_method,
            url=request.url,
            params=request.query_params,
            json=request.body,
            headers=request.headers,
        ) as response:
            try:
                data = await response.json()
            except ClientError as e:
                raise ClientLibraryError from e
            except JSONDecodeError as e:
                raise MalformedResponseError from e

            return Response(
                data=data,
                status_code=response.status,
            )

    @override
    async def close(self) -> None:
        await self._session.close()



from collections.abc import MutableMapping
from typing import Any, override

from pydantic import TypeAdapter

from retejo.integrations.common.base import BaseClient, MarkersFactorties
from retejo.interfaces import Factory
from retejo.markers import (
    BodyMarker,
    HeaderMarker,
    QueryParamMarker,
    UrlVarMarker,
)


class PydanticFactory(Factory):
    _cache_type_adapters: MutableMapping[Any, TypeAdapter]

    def __init__(self) -> None:
        self._cache_type_adapters = {}

    def _get_type_adapter(self, tp: Any) -> TypeAdapter:
        type_adapter_cache = self._cache_type_adapters.get(tp)
        if type_adapter_cache is not None:
            return type_adapter_cache

        type_adapter = TypeAdapter(tp)
        self._cache_type_adapters[tp] = type_adapter
        return type_adapter

    @override
    def load(self, data: Any, tp: Any, /) -> Any:
        type_adapter = self._get_type_adapter(tp)
        return type_adapter.validate_python(data)

    @override
    def dump(self, data: Any, tp: Any | None = None, /) -> Any:
        type_adapter = self._get_type_adapter(tp)
        return type_adapter.dump_python(data, exclude_unset=True)


class BasePydanticClient(BaseClient):
    @override
    def init_markers_factories(self) -> MarkersFactorties:
        factory = PydanticFactory()

        return {
            BodyMarker: factory,
            UrlVarMarker: factory,
            HeaderMarker: factory,
            QueryParamMarker: factory,
        }

    @override
    def init_response_factory(self) -> Factory:
        return PydanticFactory()



from abc import abstractmethod
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Protocol, runtime_checkable

from retejo._internal.interfaces.request_context_builder import RequestContext


@dataclass(slots=True, frozen=True)
class Request:
    url: str
    http_method: str
    context: RequestContext
    body: Mapping[str, str] | None = None
    headers: Mapping[str, str] | None = None
    query_params: Mapping[str, str] | None = None


@dataclass(frozen=True, slots=True)
class Response:
    data: Mapping[str, Any]
    status_code: int


@runtime_checkable
class AsyncSendableRequest(Protocol):
    @abstractmethod
    async def send_request(
        self,
        request: Request,
    ) -> Response:
        raise NotImplementedError


@runtime_checkable
class SyncSendableRequest(Protocol):
    @abstractmethod
    def send_request(
        self,
        request: Request,
    ) -> Response:
        raise NotImplementedError



from typing import Annotated, Any, TypeGuard, get_args, get_origin

from retejo._internal.markers.base import BaseMarker


class Omitted(BaseMarker):
    def __bool__(self) -> bool | None:
        return False


type Omittable[T] = T | Omitted


def is_omitted(obj: Any) -> TypeGuard[Omitted]:
    return isinstance(obj, Omitted)


def is_omittable(tp: Any) -> TypeGuard[Omittable[Any]]:
    origin = get_origin(tp)

    if origin is Annotated:
        tp = get_args(tp)[0]

    return get_origin(tp) is Omittable



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, dataclass_transform, get_args, get_origin

from retejo._internal.method.context import MethodContext, create_method_context
from retejo._internal.parents_resolver import ParentsResolver


def get_returning_tp(tp: Any) -> Any:
    parents = ParentsResolver().get_parents(tp)

    for parent in parents:
        if get_origin(parent) is Method:  # type: ignore[comparison-overlap]
            return get_args(parent)[0]

    raise RuntimeError(f"Not found __returning__ type by {tp!r} type")


@dataclass_transform(frozen_default=True)
class MethodMetaClass(type):
    def __new__(cls, name: str, bases: tuple[type, ...], namespace: dict[str, Any]) -> Any:
        klass: Any = type.__new__(cls, name, bases, namespace)

        if klass.__name__ == "Method":
            return klass

        klass = dataclass(frozen=True)(klass)

        klass.__returning__ = get_returning_tp(klass)
        klass.__context__ = create_method_context(klass)

        return klass


class Method[T](metaclass=MethodMetaClass):
    @property
    @abstractmethod
    def __url__(self) -> str:
        raise NotImplementedError

    @property
    @abstractmethod
    def __method__(self) -> str:
        raise NotImplementedError

    # fill in meta class
    @property
    def __returning__(self) -> type[T]:
        raise NotImplementedError

    @property
    def __context__(self) -> MethodContext:
        raise NotImplementedError



from collections.abc import Mapping
from typing import Any

from requests import Session

from retejo.integrations.adaptix.base import BaseAdaptixClient
from retejo.integrations.common.requests import RequestsBaseClient


class RequestsAdaptixClient(
    BaseAdaptixClient,
    RequestsBaseClient,
):
    def __init__(
        self,
        base_url: str,
        session: Session | None = None,
        cookies: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
    ) -> None:
        super().__init__(
            base_url=base_url,
            session=session,
            cookies=cookies,
            headers=headers,
        )



from abc import abstractmethod
from typing import Protocol, runtime_checkable

from retejo._internal.interfaces.sendable_method import AsyncSendableMethod, SyncSendableMethod
from retejo._internal.interfaces.sendable_request import (
    AsyncSendableRequest,
    Response,
    SyncSendableRequest,
)


@runtime_checkable
class SyncClient(
    SyncSendableRequest,
    SyncSendableMethod,
    Protocol,
):
    @abstractmethod
    def _handle_error_response(self, response: Response) -> None:
        raise NotImplementedError

    @abstractmethod
    def _handle_response(self, response: Response) -> None:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None:
        raise NotImplementedError


@runtime_checkable
class AsyncClient(
    AsyncSendableRequest,
    AsyncSendableMethod,
    Protocol,
):
    @abstractmethod
    async def _handle_error_response(self, response: Response) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _handle_response(self, response: Response) -> None:
        raise NotImplementedError

    @abstractmethod
    async def close(self) -> None:
        raise NotImplementedError

    async def aclose(self) -> None:
        await self.close()



from retejo._internal.markers.base import BaseMarker, get_marker_type, is_marker
from retejo._internal.markers.body import Body, BodyMarker, is_body
from retejo._internal.markers.file import File, FileMarker, is_file
from retejo._internal.markers.header import Header, HeaderMarker, is_header
from retejo._internal.markers.omitted import Omittable, Omitted, is_omittable, is_omitted
from retejo._internal.markers.query_param import QueryParam, QueryParamMarker, is_query_param
from retejo._internal.markers.url_var import UrlVar, UrlVarMarker, is_url_var

__all__ = (
    "BaseMarker",
    "Body",
    "BodyMarker",
    "File",
    "FileMarker",
    "Header",
    "HeaderMarker",
    "Omittable",
    "Omitted",
    "QueryParam",
    "QueryParamMarker",
    "UrlVar",
    "UrlVarMarker",
    "get_marker_type",
    "is_body",
    "is_file",
    "is_header",
    "is_marker",
    "is_omittable",
    "is_omitted",
    "is_query_param",
    "is_url_var",
)



# https://github.com/reagento/dishka/blob/develop/src/dishka/entities/with_parents.py

import typing
from abc import ABC, ABCMeta
from collections.abc import Iterable
from enum import Enum
from itertools import chain
from typing import Final, Generic, Protocol

from retejo._internal._adaptix.common import TypeHint
from retejo._internal._adaptix.feature_requirement import (
    HAS_TV_TUPLE,
    HAS_UNPACK,
)
from retejo._internal._adaptix.type_tools import (  # type: ignore[attr-defined]
    normalize_type,
)
from retejo._internal._adaptix.type_tools.basic_utils import (
    get_type_vars_of_parametrized,
    is_generic,
    is_parametrized,
)
from retejo._internal._adaptix.type_tools.fundamentals import (
    get_generic_args,
    get_type_vars,
    strip_alias,
)
from retejo._internal._adaptix.type_tools.implicit_params import fill_implicit_params

IGNORE_TYPES: Final = (
    type,
    object,
    Enum,
    ABC,
    ABCMeta,
    Generic,
    Protocol,
    Exception,
    BaseException,
)

type TypeArgsTuple = tuple[TypeHint, ...]


def is_ignored_type(origin_type: TypeHint) -> bool:
    return origin_type in IGNORE_TYPES


class ParentsResolver:
    def get_parents(self, child_type: TypeHint) -> list[TypeHint]:
        if is_ignored_type(strip_alias(child_type)):
            raise ValueError(f"The starting class {child_type!r} is in ignored type")
        return list(self._resolve_parents(child_type))

    def _resolve_parents(self, tp: TypeHint) -> Iterable[TypeHint]:
        result = [tp]
        for parent in self._fetch_parents(tp):
            if not is_ignored_type(parent):
                result.extend(self._resolve_parents(parent))
        return result

    def _fetch_parents(self, tp: TypeHint) -> list[TypeHint]:
        if is_parametrized(tp):
            return self._get_parents_of_parametrized_generic(tp)
        if is_generic(tp):
            return self._get_parents_of_parametrized_generic(
                fill_implicit_params(tp),
            )
        return self._get_parents(tp)

    def _get_parents_of_parametrized_generic(
        self,
        parametrized_generic: TypeHint,
    ) -> list[TypeHint]:
        origin = strip_alias(parametrized_generic)
        type_var_to_actual = self._get_type_var_to_actual(
            get_type_vars(origin),
            self._unpack_args(get_generic_args(parametrized_generic)),
        )
        return [
            self._parametrize_by_dict(type_var_to_actual, tp) for tp in self._get_parents(origin)
        ]

    def _unpack_args(self, args: TypeArgsTuple) -> TypeArgsTuple:
        if HAS_UNPACK and any(strip_alias(arg) == typing.Unpack for arg in args):
            subscribed = tuple[args]  # type: ignore[valid-type]
            return tuple(arg.source for arg in normalize_type(subscribed).args)
        return args

    def _get_type_var_to_actual(
        self,
        type_vars: TypeArgsTuple,
        args: TypeArgsTuple,
    ) -> dict[TypeHint, TypeArgsTuple]:
        result = {}
        idx = 0
        for tv in type_vars:
            if HAS_TV_TUPLE and isinstance(tv, typing.TypeVarTuple):  # type: ignore[attr-defined, unused-ignore]
                tuple_len = len(args) - len(type_vars) + 1
                result[tv] = args[idx : idx + tuple_len]
                idx += tuple_len
            else:
                result[tv] = (args[idx],)
                idx += 1

        return result

    def _parametrize_by_dict(
        self,
        type_var_to_actual: dict[TypeHint, TypeArgsTuple],
        tp: TypeHint,
    ) -> TypeHint:
        params = get_type_vars_of_parametrized(tp)
        if not params:
            return tp
        return tp[
            tuple(
                chain.from_iterable(type_var_to_actual[type_var] for type_var in params),
            )
        ]

    def _get_parents(self, tp: TypeHint) -> list[TypeHint]:
        if hasattr(tp, "__orig_bases__"):
            return [
                parent
                for parent in tp.__orig_bases__
                if strip_alias(parent) not in (Generic, Protocol)
            ]
        return list(tp.__bases__)



from typing import override

from adaptix import Retort, as_sentinel, name_mapping

from retejo.integrations.common.base import BaseClient, MarkersFactorties
from retejo.interfaces import Factory
from retejo.markers import (
    BodyMarker,
    HeaderMarker,
    Omitted,
    QueryParamMarker,
    UrlVarMarker,
)


class BaseAdaptixClient(BaseClient):
    @override
    def init_markers_factories(self) -> MarkersFactorties:
        retort = Retort(
            recipe=[
                as_sentinel(Omitted),
                name_mapping(
                    omit_default=True,
                ),
            ],
        )

        return {
            BodyMarker: retort,
            UrlVarMarker: retort,
            HeaderMarker: retort,
            QueryParamMarker: retort,
        }

    @override
    def init_response_factory(self) -> Factory:
        return Retort()



from .aiohttp import AiohttpAdaptixClient
from .base import BaseAdaptixClient
from .requests import RequestsAdaptixClient

__all__ = [
    "AiohttpAdaptixClient",
    "BaseAdaptixClient",
    "RequestsAdaptixClient",
]



from collections.abc import Callable

from retejo._internal.bind_method import _BindMethod
from retejo.method import Method


def bind_method[**P, T](method: Callable[P, Method[T]]) -> _BindMethod[P, T]:
    return _BindMethod(method)



from abc import abstractmethod
from typing import Protocol, runtime_checkable

from retejo.method import Method


@runtime_checkable
class AsyncSendableMethod(Protocol):
    @abstractmethod
    async def send_method[T](
        self,
        method: Method[T],
    ) -> T:
        raise NotImplementedError


@runtime_checkable
class SyncSendableMethod(Protocol):
    @abstractmethod
    def send_method[T](
        self,
        method: Method[T],
    ) -> T:
        raise NotImplementedError



from collections.abc import Awaitable, Callable
from typing import Any, NoReturn, overload

from retejo.interfaces import AsyncSendableMethod, SyncSendableMethod
from retejo.method import Method


class _BindMethod[**P, T]:
    def __init__(
        self,
        method: Callable[P, Method[T]],
    ) -> None:
        self._method = method

    @overload
    def __get__(self, obj: SyncSendableMethod, objtype: Any = None) -> Callable[P, T]: ...

    @overload
    def __get__(
        self, obj: AsyncSendableMethod, objtype: Any = None
    ) -> Callable[P, Awaitable[T]]: ...

    @overload
    def __get__(self, obj: Any, objtype: Any = None) -> NoReturn: ...

    def __get__(self, obj: Any, objtype: Any = None) -> Any:
        if isinstance(obj, SyncSendableMethod):

            def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return obj.send_method(self._method(*args, **kwargs))

            return sync_wrapper

        if isinstance(obj, AsyncSendableMethod):

            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
                return await obj.send_method(self._method(*args, **kwargs))

            return async_wrapper

        raise RuntimeError("method_call use is only Session subclasses")
